#include <iostream>
#include <stdio.h>  
#include <opencv2/opencv.hpp>  
#include <vector>  
#include <fstream>

#define SQRT2 1.41
#define SQRT3 1.7321

struct normalizedData {
	cv::Mat T2D;
	cv::Mat T3D;
	std::vector<cv::Point2f> newPts2D;
	std::vector<cv::Point3f> newPts3D;
};

normalizedData NormalizeData(std::vector<cv::Point3f> pts3D, std::vector<cv::Point2f> pts2D) {
	int ptsNum = pts3D.size();

	//calculate means (they will be the center of coordinate systems)
	cv::Point2f mean1(0, 0);
	cv::Point3f mean2(0, 0, 0);

	for (int i = 0; i < ptsNum; i++) {
		mean1 += pts2D[i];
		mean2 += pts3D[i];
	}
	
	mean1 /= (float)ptsNum;
	mean2 /= (float)ptsNum;

	float spread1 = 0.0;
	float spread2 = 0.0;

	for (int i = 0; i < ptsNum; i++) {
		spread1 += cv::norm(pts2D[i] - mean1);
		spread2 += cv::norm(pts3D[i] - mean2);
	}

	spread1 /= (float)ptsNum;
	spread2 /= (float)ptsNum;

	cv::Mat offs1 = cv::Mat::eye(3, 3, CV_32F);
	cv::Mat offs2 = cv::Mat::eye(4, 4, CV_32F);
	cv::Mat scale1 = cv::Mat::eye(3, 3, CV_32F);
	cv::Mat scale2 = cv::Mat::eye(4, 4, CV_32F);

	offs1.at<float>(0, 2) = -mean1.x;
	offs1.at<float>(1, 2) = -mean1.y;

	offs2.at<float>(0, 3) = -mean2.x;
	offs2.at<float>(1, 3) = -mean2.y;
	offs2.at<float>(2, 3) = -mean2.z;

	scale1.at<float>(0, 0) = SQRT2 / spread1;
	scale1.at<float>(1, 1) = SQRT2 / spread1;

	scale2.at<float>(0, 0) = SQRT3 / spread2;
	scale2.at<float>(1, 1) = SQRT3 / spread2;
	scale2.at<float>(2, 2) = SQRT3 / spread2;

	normalizedData ret;
	ret.T2D = scale1 * offs1;
	ret.T3D = scale2 * offs2;

	for (int i = 0; i < ptsNum; i++) {
		cv::Point2f p2D;
		cv::Point3f p3D;

		p2D.x = SQRT2 * (pts2D[i].x - mean1.x) / spread1;
		p2D.y = SQRT2 * (pts2D[i].y - mean1.y) / spread1;

		p3D.x = SQRT3 * (pts3D[i].x - mean2.x) / spread2;
		p3D.y = SQRT3 * (pts3D[i].y - mean2.y) / spread2;
		p3D.z = SQRT3 * (pts3D[i].z - mean2.z) / spread2;

		ret.newPts2D.push_back(p2D);
		ret.newPts3D.push_back(p3D);
	}

	return ret;

}

cv::Mat calcProjMtx(std::vector<cv::Point3f> objPts, std::vector<cv::Point2f> imgPts) {
	const int ptsNum = objPts.size();
	normalizedData nd = NormalizeData(objPts, imgPts);

	cv::Mat A(2 * ptsNum, 12, CV_32F);

	for (int i = 0; i < ptsNum; i++) {

		float X = nd.newPts3D[i].x;
		float Y = nd.newPts3D[i].y;
		float Z = nd.newPts3D[i].z;

		float u = nd.newPts2D[i].x;
		float v = nd.newPts2D[i].y;

		A.at<float>(2 * i, 0) = X;
		A.at<float>(2 * i, 1) = Y;
		A.at<float>(2 * i, 2) = Z;
		A.at<float>(2 * i, 3) = 1.0f;
		A.at<float>(2 * i, 4) = 0.0f;
		A.at<float>(2 * i, 5) = 0.0f;
		A.at<float>(2 * i, 6) = 0.0f;
		A.at<float>(2 * i, 7) = 0.0f;
		A.at<float>(2 * i, 8) = -u * X;
		A.at<float>(2 * i, 9) = -u * Y;
		A.at<float>(2 * i, 10) = -u * Z;
		A.at<float>(2 * i, 11) = -u;

		A.at<float>(2 * i + 1, 0) = 0.0f;
		A.at<float>(2 * i + 1, 1) = 0.0f;
		A.at<float>(2 * i + 1, 2) = 0.0f;
		A.at<float>(2 * i + 1, 3) = 0.0f;
		A.at<float>(2 * i + 1, 4) = X;
		A.at<float>(2 * i + 1, 5) = Y;
		A.at<float>(2 * i + 1, 6) = Z;
		A.at<float>(2 * i + 1, 7) = 1.0f;
		A.at<float>(2 * i + 1, 8) = -v * X;
		A.at<float>(2 * i + 1, 9) = -v * Y;
		A.at<float>(2 * i + 1, 10) = -v * Z;
		A.at<float>(2 * i + 1, 11) = -v;


	}

	cv::Mat eVecs(12, 12, CV_32F), eVals(12, 12, CV_32F);
	eigen(A.t() * A, eVals, eVecs);

	cv::Mat P(3, 4, CV_32F);
	for (int i = 0; i < 12; i++) P.at<float>(i / 4, i % 4) = eVecs.at<float>(11, i);


	//For normalization:
	P = (nd.T2D.inv()) * P * nd.T3D;

	std::cout << "T2D:\n" << nd.T2D << std::endl;
	std::cout << "T3D:\n" << nd.T3D << std::endl;


	//Normalize:
	P = P * (1.0 / P.at<float>(2, 3));
	std::cout << P << std::endl;

	return P;
}

std::vector<cv::Point2f> readImagePoints(std::string fn)
{
	std::vector<cv::Point2f> points;
	std::ifstream f(fn);

	while (!f.eof())
	{
		float x, y;
		f >> x >> y;
		points.push_back({ x,y });
	}

	f.close();
	return points;
}

std::vector<cv::Point3f> readWorldPoints(std::string fn)
{
	std::vector<cv::Point3f> points;
	std::ifstream f(fn);

	while (!f.eof())
	{
		float x, y, z;
		f >> x >> y >> z;
		points.push_back({ x,y,z });
	}

	f.close();
	return points;
}

void calibrateWithCube()
{
	// Create a synthetic projection matrix
	std::vector<cv::Point3f> pts3D = readWorldPoints("world_points.xyz");
	std::vector<cv::Point2f> pts2D = readImagePoints("image_points.xy");

	cv::Mat img = cv::imread("cube.jpg");
	for (auto p : pts2D)
	{
		cv::circle(img, p, 5, { 0,0,255 }, cv::FILLED);
	}
	cv::imshow("Calibration", img);
	cv::waitKey(0);

	cv::Mat intrinsic_Matrix(3, 3, CV_64F);

	cv::Mat P = calcProjMtx(pts3D, pts2D);

	cv::Mat K(3, 3, CV_32F); // intrinsic parameter matrix
	cv::Mat R(3, 3, CV_32F); // rotation matrix
	cv::Mat T(4, 1, CV_32F); // translation vector

	cv::decomposeProjectionMatrix(P, K, R, T);
	K = K * (1.0 / K.at<float>(2, 2));
	T = T * (1.0 / T.at<float>(3));
	std::cout << "K:\n" << K << std::endl;
	std::cout << "R:\n" << R << std::endl;
	std::cout << "T:\n" << T << std::endl;
}

void calibrateWithChessboard()
{
	cv::Size boardSize{ 8, 6 };
	cv::Size imgSize;

	std::vector<std::vector<cv::Point3f>> objPoints;
	std::vector<std::vector<cv::Point2f>> imgPoints;

	std::vector<cv::Point3f> tmp_objPoints;
	for (float i = 0.0f; i < boardSize.height; i += 1.0f) {
		for (float j = 0.0f; j < boardSize.width; j += 1.0f) {
			tmp_objPoints.push_back({ i, j, 0.0f });
		}
	}

	for (int i = 0; i <= 5; ++i)
	{
		std::string fname = "images/image_" + std::to_string(i) + ".jpg";
		cv::Mat img = cv::imread(fname, cv::IMREAD_COLOR);
		imgSize = img.size();

		cv::Mat gray;
		cv::cvtColor(img, gray, cv::COLOR_BGR2GRAY);

		std::vector<cv::Point2f> corners;
		bool found = cv::findChessboardCorners(gray, boardSize, corners);
		if (found) {
			cv::cornerSubPix(gray, corners, cv::Size(7, 7), cv::Size(-1, -1), cv::TermCriteria(cv::TermCriteria::EPS + cv::TermCriteria::COUNT, 300, 0.00001));

			cv::drawChessboardCorners(img, boardSize, corners, found);
			cv::imshow("Chessboard", img);
			cv::waitKey(0);

			imgPoints.push_back(corners);
			objPoints.push_back(tmp_objPoints);
		}
		else {
			std::cout << "Not found!" << std::endl;
		}
	}

	cv::Mat K;
	cv::Mat distCoeffs;
	std::vector<cv::Mat> rvecs, tvecs;
	float error = cv::calibrateCamera(objPoints, imgPoints, imgSize, K, distCoeffs, rvecs, tvecs);
	std::cout << "Overall error: " << error << std::endl;

	std::cout << K << std::endl;

	std::cout << distCoeffs << std::endl;

	std::cout << rvecs[0] << std::endl;
	std::cout << tvecs[0] << std::endl;

	for (int i = 0; i < 5; ++i) {
		float mean_reprojection_error = 0.0f;
		std::vector<cv::Point2f> reProjectedPoints;

		cv::projectPoints(tmp_objPoints, rvecs[i], tvecs[i], K, cv::noArray(), reProjectedPoints);
		for (int j = 0; j < reProjectedPoints.size(); ++j) {
			mean_reprojection_error += cv::norm(imgPoints[i][j] - reProjectedPoints[j]);
		}

		mean_reprojection_error /= reProjectedPoints.size();
		std::cout << "Reprojection error: " << mean_reprojection_error << std::endl;
	}

	for (int i = 0; i <= 5; ++i)
	{
		std::string fname = "images/image_" + std::to_string(i) + ".jpg";
		cv::Mat img = cv::imread(fname, cv::IMREAD_COLOR);
		cv::Mat new_img;
		cv::undistort(img, new_img, K, distCoeffs);
		cv::imshow("Before", img);
		cv::imshow("After", new_img);
		cv::waitKey(0);
	}
}

int main(int argc, char* argv[])
{
	calibrateWithCube();
	//calibrateWithChessboard();

	return 0;
}
